/* gkrellmoon.c 
 * Copyright (C) 2001, Dale P. Smith, Altus Technologies Corporation
 */

#include <gkrellm/gkrellm.h>

#include "moon_60.xpm"

#define HEIGHT 49
#define IMAGE_WIDTH 48
#define IMAGE_HEIGHT 48
#define IMAGE_COUNT 60


#define STYLE_NAME "moon"
#define PLUGIN_CONFIG_KEYWORD    "moon"
#define MOONCLOCK_MAJOR_VERSION	0
#define MOONCLOCK_MINOR_VERSION 3

#undef FORMAT_TIMES

typedef struct {
    int longitude;
    int latitude;
    int testing;
} Options;

static Panel   *panel;
static Options options;
static gint    style_id;

static GtkWidget *longitude_spin_button;
static GtkWidget *latitude_spin_button;
static GtkWidget *testing_button;
static GdkPixmap *moon_image = NULL;
static GdkBitmap *moon_mask = NULL;
static Decal     *moon = NULL;

static GtkTooltips	*tooltip;
#include "CalcEphem.h"
#include "MoonRise.h"

typedef struct CTrans MoonData;

typedef struct _Moon Moon;
struct _Moon {
    MoonData data;
};

static Moon moonunit;

static void update_tooltip(Moon *moon);

static void
update_moon_data(Moon * moon)
{
    struct tm *time_struc;	/* The tm struct is defined in <time.h> */
    gdouble local_std_time, univ_time, eot;
    glong current_gmt, date;
    gint day_of_month, month, year;

    current_gmt = time(CurrentTime);	/* CurrentTime defined in <X11/X.h> */

    time_struc = gmtime(&current_gmt);
    univ_time =
	time_struc->tm_hour + time_struc->tm_min / 60.0 +
	time_struc->tm_sec / 3600.0;

    /* The date needs to be the date in UTC, i.e. in greenwich, so
     * be sure not to call the localtime function until after date
     * has been set (there's only one tm structure).  */

    year = time_struc->tm_year + 1900;
    month = time_struc->tm_mon + 1;
    day_of_month = time_struc->tm_mday;

    date = year * 10000 + month * 100 + day_of_month;

    time_struc = localtime(&current_gmt);
    local_std_time =
	time_struc->tm_hour + time_struc->tm_min / 60.0 +
	time_struc->tm_sec / 3600.0;

    moon->data.Glat = options.latitude;
    moon->data.Glon = options.longitude;

    CalcEphem(date, univ_time, &moon->data);

    moon->data.LST = local_std_time;
    moon->data.LMT = univ_time - moon->data.Glon / 15.0;
    if (moon->data.LMT < 0.0)
	moon->data.LMT += 24.0;
    if (moon->data.LMT > 24.0)
	moon->data.LMT -= 24.0;

    /* eot is the equation of time. gmst is Greenwich Sidereal
     * Time.  This equation below is correct, but confusing at
     * first.  It's easy to see when you draw the following
     * picture: A sphere with 0 and 180 degree longitude, North on
     * top, a meridian for the real sun, a meridian for a fictive
     * average sun, a meridian denoting the vernal equinox.  Note
     * that universal time is the hour angle between 180 degrees
     * and the fictive sun's meridian measured clockwise.  gmst is
     * the hour angle between 0 degrees and the meridian of the
     * vernal equinox measured clockwise.  RA_sun/15.0 is the hour
     * angle of the real sun measured counterclockwise from the
     * vernal equinox. eot is the difference between the real and
     * the fictive sun.  Looking at the picture, it's easy to see
     * that 12=RA_sun/15-gmst+eot+utc (12 hours = 180 deg.) */

    eot =
	12.0 - univ_time + moon->data.gmst - moon->data.RA_sun / 15.0;

    if (eot < 0.0)
	eot += 24.0;
    if (eot > 24.0)
	eot -= 24.0;

    moon->data.LAT = moon->data.LMT + eot;
    if (moon->data.LAT < 0.0)
	moon->data.LAT += 24.0;
    if (moon->data.LAT > 24.0)
	moon->data.LAT -= 24.0;

    update_tooltip(moon);
}


static int
moon_image_number(Moon * moon)
{
    gdouble image_float;
    gint image_int;
    gint image_number;

    /* MoonPhase expresses phase of moon as fraction of 1; 0.5=full. */
    image_float = moon->data.MoonPhase * (gdouble) 60;
    image_int = (gint) image_float;

    if ((image_float - image_int) >= 0.5)
	image_number = (image_int + 1) % 60;
    else
	image_number = image_int % 60;

    return image_number;
}


static void
moon_update_plugin()
{
    static int image_number;

    /* Draw plugin specific data on the chart */
    /* Use xlib or gdk functions, or gkrellm_draw_chart() if applicable */

    if (GK.minute_tick) {
	update_moon_data(&moonunit);
    }
	
    if (options.testing) {
	++image_number;
	image_number %= IMAGE_COUNT;
    } else {
	image_number = moon_image_number(&moonunit);
    }

    gkrellm_draw_decal_pixmap(panel, moon, image_number);
    gkrellm_draw_layers(panel);
}

#ifdef FORMAT_TIMES
static void
format_time(gdouble time)
{
    gint hour, min;

    hour = (gint) (time);
    min = (gint) ((ABS(time - hour)) * 60);
    printf("%02d:%02d\n", hour, min);
}

static void
format_angle_hours(gdouble angle)
{
    gint hour, min;

    hour = (gint) (angle / 15.0);
    min = (gint) ((ABS(angle - hour * 15.0)) * 4.0);

    printf("%02d:%02d\n", hour, min);
}


static void
calc_riseset_time(CTrans * c)
{
    gdouble rise, set;

    MoonRise(c, &rise, &set);

    if (abs(rise) > 24.0)
	printf("no rise");
    else
	format_time(rise);

    if (abs(set) > 24.0)
	printf("no set");
    else
	format_time(set);
}
#endif

static void
update_tooltip(Moon *moon)
{
    GString	*mboxes = NULL;
    gchar	buf[128];

    if (tooltip == NULL)
	return;

#ifdef FORMAT_TIMES
    calc_riseset_time(&moon->data);
    moon->data.day -= 1;
    calc_riseset_time(&moon->data);
    moon->data.day += 2;
    calc_riseset_time(&moon->data);
    moon->data.day -= 1;
#endif

#if 0
    printf("A: "); format_time(moon->data.A_moon);
    printf("h: "); format_time(moon->data.h_moon);
    printf("EarthMoon Distance: %5.1f Re\n", moon->data.EarthMoonDistance);
#endif

#if 0
    printf("RA: "); format_angle_hours(moon->data.RA_moon);
    printf("Declination: "); format_time(moon->data.DEC_moon);
    printf("EarthMoon Distance: %5.1f Re\n", moon->data.EarthMoonDistance);
    printf("\n");
#endif

    mboxes = g_string_sized_new(512);
    snprintf(buf, sizeof(buf), "Age: %2.2f Days\n",
	     moon->data.MoonAge);
    g_string_append(mboxes, buf);
    
    snprintf(buf, sizeof(buf), "Frac: %5.1f%%\n",
	     100 * moon->data.MoonPhase);
    g_string_append(mboxes, buf);
    
    snprintf(buf, sizeof(buf), "Illum: %5.1f%%\n",
	     50.0 * (1.0 - cos(moon->data.MoonPhase * 6.2831853)));
    g_string_append(mboxes, buf);
    
    snprintf(buf, sizeof(buf), "Visible: %s\n",
	     moon->data.Visible ? "Yes" : "No");
    g_string_append(mboxes, buf);
    
    gtk_tooltips_set_tip(tooltip, panel->drawing_area, mboxes->str, NULL);
    gtk_tooltips_set_delay(tooltip, 750);
    gtk_tooltips_enable(tooltip);
    if (mboxes)
	g_string_free(mboxes, TRUE);
}


static gint
panel_expose_event(GtkWidget *widget, GdkEventExpose *ev)
{
    gdk_draw_pixmap(widget->window,
		    widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
		    panel->pixmap, ev->area.x, ev->area.y, ev->area.x, ev->area.y,
		    ev->area.width, ev->area.height);
    return FALSE;
}

static void
load_images()
{
    GdkImlibImage  *image = NULL;

    gkrellm_load_image(NULL, moon_60_xpm, &image, NULL);
    gkrellm_render_to_pixmap(image, &moon_image, &moon_mask, 0, 0);
}

static void
moon_create_plugin(GtkWidget *vbox, gint first_create)
{
    Style	   *style = NULL;
    int            image_x_offset;
    int            image_y_offset;

    load_images();

    if (first_create)
        panel = gkrellm_panel_new0();
    else {
	gkrellm_destroy_decal_list(panel);
    }

    style = gkrellm_meter_style(style_id);

    image_x_offset = (gkrellm_chart_width() - IMAGE_WIDTH) / 2;
    image_y_offset = (HEIGHT - IMAGE_HEIGHT) / 2;

    moon = gkrellm_create_decal_pixmap(panel, moon_image, moon_mask,
				       IMAGE_COUNT,
				       style, image_x_offset, image_y_offset);


    panel->textstyle = gkrellm_meter_textstyle(style_id);
    panel->label->h_panel = HEIGHT;      /* Whatever height you need */
    gkrellm_create_panel(vbox, panel, gkrellm_bg_meter_image(style_id));
    gkrellm_monitor_height_adjust(panel->h);


    if (first_create) {
        gtk_signal_connect(GTK_OBJECT(panel->drawing_area),
			   "expose_event", (GtkSignalFunc) panel_expose_event,
			   NULL);
	tooltip=gtk_tooltips_new();
    }
    update_moon_data(&moonunit);
    gkrellm_draw_decal_pixmap(panel, moon, moon_image_number(&moonunit));
    gkrellm_draw_layers(panel);
}


static void
moon_create_tab(GtkWidget *tab_vbox)
{
    GtkWidget		*tabs;
    GtkWidget		*vbox;

    tabs = gtk_notebook_new();
    gtk_notebook_set_tab_pos(GTK_NOTEBOOK(tabs), GTK_POS_TOP);
    gtk_box_pack_start(GTK_BOX(tab_vbox), tabs, TRUE, TRUE, 0);
    
/* --Setup Tab */
    vbox = gkrellm_create_tab(tabs, _("Setup"));

    gkrellm_spin_button(vbox, &longitude_spin_button,
			(gfloat) options.longitude,
			-180.0, 180.0, 1.0, 1.0, 0, 60, NULL, NULL,
			FALSE, _("Longitude (decimal degrees + = W, - = E)"));

    gkrellm_spin_button(vbox, &latitude_spin_button,
			(gfloat) options.latitude,
			-90.0, 90.0, 1.0, 1.0, 0, 60, NULL, NULL,
			FALSE, _("Latitude (decimal degrees + = N, - = S)"));

    gkrellm_check_button(vbox, &testing_button,
			options.testing,
			TRUE, 0, _("Spin images fast"));

#if 0
/* --Info tab */
	vbox = gkrellm_create_tab(tabs, _("Info"));
	scrolled = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled),
			GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_box_pack_start(GTK_BOX(vbox), scrolled, TRUE, TRUE, 0);
	text = gtk_text_new(NULL, NULL);
	for (i = 0; i < sizeof(moon_info_text)/sizeof(gchar *); ++i)
		gkrellm_add_info_text_string(text, _(mail_info_text[i]));
	gtk_text_set_editable(GTK_TEXT(text), FALSE);
	gtk_container_add(GTK_CONTAINER(scrolled), text);
#endif

/* ----------------- about text --------------------*/

	{
	    gchar *plugin_about_text;
	    GtkWidget *label, *text;

	    plugin_about_text = g_strdup_printf(
		"MoonClock %d.%d\n"
		"GKrellM MoonClock Plugin\n\n"
		"Copyright (C) 2001 Dale P. Smith\n"
		"dsmith@altustech.com\n\n"
		"Released under the GNU Public Licence",
		MOONCLOCK_MAJOR_VERSION, MOONCLOCK_MINOR_VERSION);
	    
	    text = gtk_label_new(plugin_about_text); 
	    label = gtk_label_new("About");
	    gtk_notebook_append_page(GTK_NOTEBOOK(tabs),text,label);
	    g_free(plugin_about_text);
	}
}

static void
moon_apply_config()
{
    options.longitude = gtk_spin_button_get_value_as_int(
	GTK_SPIN_BUTTON(longitude_spin_button));

    options.latitude = gtk_spin_button_get_value_as_int(
	GTK_SPIN_BUTTON(latitude_spin_button));

    options.testing = GTK_TOGGLE_BUTTON(testing_button)->active;
}

static void
moon_save_config (FILE *f)
{
    fprintf(f, "%s longitude %d\n", PLUGIN_CONFIG_KEYWORD, options.longitude);
    fprintf(f, "%s latitude %d\n", PLUGIN_CONFIG_KEYWORD, options.latitude);
}

static void
moon_load_config (gchar *arg)
{
    gchar config[64], item[256];
    gint n;

    n = sscanf(arg, "%s %[^\n]", config, item);
    if (n != 2)
	return;

    if (strcmp(config, "longitude") == 0)
	sscanf(item, "%d\n", &(options.longitude));
    if (strcmp(config, "latitude") == 0)
	sscanf(item, "%d\n", &(options.latitude));
}

static Monitor  plugin_mon  = {
    "Moon Clock",		/* Name, for config tab.        */
    0,				/* Id,  0 if a plugin           */
    moon_create_plugin,		/* The create_plugin() function */
    moon_update_plugin,		/* The update_plugin() function */
    moon_create_tab,		/* The create_plugin_tab() config function */
    moon_apply_config,		/* The apply_plugin_config() function      */

    moon_save_config,		/* The save_plugin_config() function  */
    moon_load_config,		/* The load_plugin_config() function  */
    PLUGIN_CONFIG_KEYWORD,	/* config keyword                     */

    NULL,			/* Undefined 2  */
    NULL,			/* Undefined 1  */
    NULL,			/* private		*/

    MON_INSERT_AFTER|MON_CLOCK,	/* Insert plugin before this monitor.       */
    NULL,			/* Handle if a plugin, filled in by GKrellM */
    NULL			/* path if a plugin, filled in by GKrellM   */
};

Monitor *
init_plugin(void)
{
    options.longitude = 81;	/* Where I live! */
    options.latitude = 44;

    style_id = gkrellm_add_meter_style(&plugin_mon, STYLE_NAME);

    return &plugin_mon;
}
